import numpy as np
import numpy.testing as npt
import pytest
import znsocket
from ase.build import bulk, molecule

from zndraw import Extension, ZnDraw
from zndraw.exceptions import RoomLockedError


def run_queue(vis, key, msg: dict):
    modifier_queue = znsocket.Dict(
        r=vis.r,
        socket=vis._refresh_client,
        key=f"queue:{vis.token}:{key}",
    )
    modifier_queue.update(msg)
    vis.socket.emit("room:worker:run")
    vis.socket.sleep(10)


def test_run_selection(server, s22):
    """Test the server fixture."""
    vis = ZnDraw(url=server, token="test_token")
    vis.extend(s22)
    vis.step = 0
    vis.selection = [0]

    run_queue(vis, "selection", {"ConnectedParticles": {}})

    assert vis.selection == [0, 1, 2, 3]


def test_run_modifier(server):
    vis = ZnDraw(url=server, token="test_token")
    vis.append(molecule("H2O"))
    vis.selection = [0]
    assert len(vis) == 1
    assert len(vis[-1]) == 3

    run_queue(vis, "modifier", {"Delete": {}})

    assert len(vis) == 2
    assert len(vis[-1]) == 2


def test_register_modifier(server, s22, water):
    vis = ZnDraw(url=server, token="test_token")
    vis.extend(s22)

    class AppendWater(Extension):
        def run(self, vis, **kwargs) -> None:
            vis.append(water)
            vis.step = len(vis) - 1

    vis.register_modifier(AppendWater)

    run_queue(vis, "modifier", {"AppendWater": {}})

    assert vis.atoms == water


def test_locked(server):
    vis = ZnDraw(url=server, token="test_token")

    assert vis.locked is False
    vis.locked = True
    assert vis.locked is True


##### Tests for each available modifier #####


def test_modify_delete(server):
    vis = ZnDraw(url=server, token="test_token")
    vis.append(molecule("H2O"))
    vis.selection = [0]
    assert len(vis) == 1
    assert len(vis[-1]) == 3

    run_queue(vis, "modifier", {"Delete": {}})

    assert len(vis) == 2
    assert len(vis[-1]) == 2


def test_modify_rotate(server):
    vis = ZnDraw(url=server, token="test_token")
    vis.append(molecule("H2O"))
    vis.selection = [0, 1]
    vis.points = [[0, 0, 0], [1, 0, 0]]

    run_queue(vis, "modifier", {"Rotate": {"steps": 10}})
    vis.socket.sleep(5)

    assert len(vis) == 11
    # TODO: test that the atoms rotated correctly


def test_modify_translate(server):
    vis = ZnDraw(url=server, token="test_token")
    vis.append(molecule("H2O"))
    vis.selection = [0, 1, 2]
    vis.points = [[1, 0, 0], [0, 0, 0]]

    run_queue(vis, "modifier", {"Translate": {"steps": 10}})

    assert len(vis) == 11

    orig_pos = vis[0].positions
    npt.assert_allclose(vis[10].positions, orig_pos - np.array([1, 0, 0]))
    # spline interpolation is not an exact line
    npt.assert_allclose(vis[5].positions, orig_pos - np.array([0.5, 0, 0]), rtol=0.015)

    assert len(vis.points) == 2


def test_modify_duplicate(server):
    vis = ZnDraw(url=server, token="test_token")
    vis.append(molecule("H2O"))
    vis.selection = [0]

    run_queue(vis, "modifier", {"Duplicate": {}})

    assert len(vis) == 2
    assert len(vis[0]) == 3
    assert len(vis[1]) == 4  # one duplicated atom


def test_modify_change_type(server):
    vis = ZnDraw(url=server, token="test_token")
    vis.append(molecule("H2O"))
    vis.selection = [0]

    run_queue(vis, "modifier", {"ChangeType": {"symbol": "He"}})

    assert vis[1].symbols[0] == "He"


def test_modify_wrap(server):
    vis = ZnDraw(url=server, token="test_token")
    copper = bulk("Cu", cubic=True)
    copper.positions += 5  # shift, so wrapped is recognizable
    vis.extend([copper, copper])

    run_queue(vis, "modifier", {"Wrap": {"all": True}})

    # Wrap is an inplace modifier
    assert len(vis) == 2
    for idx in range(2):
        wrapped_atoms = vis[idx]
        wrapped_atoms.wrap()
        assert not np.allclose(vis[idx].positions, copper.positions)
        assert np.allclose(vis[idx].positions, wrapped_atoms.positions)


def test_modify_replicate(server):
    vis = ZnDraw(url=server, token="test_token")
    wurtzite = bulk("ZnO", "wurtzite", a=3.25, c=5.2)
    vis.extend([wurtzite, wurtzite])

    run_queue(vis, "modifier", {"Replicate": {"x": 2, "y": 2, "z": 2, "all": True}})
    vis.socket.sleep(5)

    # Replicate is an inplace modifier
    assert len(vis) == 2
    for idx in range(2):
        assert len(vis[idx]) == 8 * len(wurtzite)


def test_modify_AddLineParticles(server):
    vis = ZnDraw(url=server, token="test_token")
    vis.append(molecule("H2O"))
    vis.points = [[0, 0, 0], [1, 0, 0]]

    run_queue(vis, "modifier", {"AddLineParticles": {"steps": 10, "symbol": "He"}})

    assert len(vis[0]) == 3
    assert len(vis[1]) == 5
    npt.assert_allclose(vis[1].positions[3], [0, 0, 0])
    npt.assert_allclose(vis[1].positions[4], [1, 0, 0])
    assert vis[1].symbols[4] == "He"
    assert vis[1].symbols[3] == "He"


def test_modify_center(server):
    vis = ZnDraw(url=server, token="test_token")
    copper = bulk("Cu", cubic=True)
    vis.append(copper)
    vis.selection = [0]

    run_queue(vis, "modifier", {"Center": {"all": True}})

    assert np.allclose(vis[0][0].position, np.diag(vis[0].cell) / 2)
    assert not np.allclose(vis[0].positions, copper.positions)


def test_modify_RemoveAtoms(server):
    vis = ZnDraw(url=server, token="test_token")
    vis.append(molecule("H2O"))
    vis.append(molecule("H2O"))
    assert len(vis) == 2
    vis.step = 0

    run_queue(vis, "modifier", {"RemoveAtoms": {}})

    assert len(vis) == 1


def test_modified_while_locked(server):
    vis = ZnDraw(url=server, token="test_token")
    vis.append(molecule("H2O"))
    vis.append(molecule("H2O"))
    vis.append(molecule("H2O"))

    assert len(vis) == 3

    run_queue(vis, "modifier", {"RemoveAtoms": {}})

    assert len(vis) == 2
    vis.locked = True

    run_queue(vis, "modifier", {"RemoveAtoms": {}})

    assert len(vis) == 2

    with pytest.raises(RoomLockedError):
        vis.append(molecule("H2O"))

    with pytest.raises(RoomLockedError):
        vis.step = 0

    with pytest.raises(RoomLockedError):
        vis.selection = [0]

    with pytest.raises(RoomLockedError):
        vis.atoms = molecule("H2O")

    with pytest.raises(RoomLockedError):
        vis.extend([molecule("H2O")])



import os

import eventlet.wsgi

eventlet.monkey_patch()  # MUST BE THERE FOR THE TESTS TO WORK

import random

import ase.build
import ase.collections
import numpy as np
import pytest
import socketio.exceptions
from ase.calculators.singlepoint import SinglePointCalculator

from zndraw.app import create_app
from zndraw.standalone import run_celery_worker


@pytest.fixture
def server():
    port = random.randint(10000, 20000)

    os.environ["FLASK_PORT"] = str(port)
    os.environ["FLASK_STORAGE"] = "redis://localhost:6379/0"
    os.environ["FLASK_SERVER_URL"] = f"http://localhost:{port}"

    proc = run_celery_worker()

    def start_server():
        app = create_app()
        app.config["TESTING"] = True

        socketio = app.extensions["socketio"]
        try:
            socketio.run(
                app,
                host="0.0.0.0",
                port=app.config["PORT"],
            )
        finally:
            app.extensions["redis"].flushall()

    thread = eventlet.spawn(start_server)

    # wait for the server to be ready
    for _ in range(100):
        try:
            with socketio.SimpleClient() as client:
                client.connect(f"http://localhost:{port}")
                break
        except socketio.exceptions.ConnectionError:
            eventlet.sleep(0.1)
    else:
        raise TimeoutError("Server did not start in time")

    yield f"http://127.0.0.1:{port}"

    thread.kill()
    proc.kill()
    proc.wait()


@pytest.fixture
def s22() -> list[ase.Atoms]:
    """S22 dataset."""
    return list(ase.collections.s22)


@pytest.fixture
def water() -> ase.Atoms:
    """Water molecule."""
    return ase.build.molecule("H2O")


@pytest.fixture
def s22_energy_forces() -> list[ase.Atoms]:
    images = []
    for atoms in ase.collections.s22:
        calc = SinglePointCalculator(
            atoms, energy=np.random.rand(), forces=np.random.rand(len(atoms), 3)
        )
        atoms.calc = calc
        images.append(atoms)
    return images



from zndraw import ZnDraw

# def test_config_defaults(server):
#     vis = ZnDraw(url=server, token="test_token")
#     ref_config = ZnDrawConfig(vis=None)

#     vis.config == ref_config.to_dict()


def test_config_modify_arrows(server):
    room = "test_config_arrows"
    vis = ZnDraw(url=server, token=room)

    vis.config["Arrows"]["normalize"] = False
    assert vis.config["Arrows"]["normalize"] is False

    vis.config["Arrows"]["normalize"] = True
    assert vis.config["Arrows"]["normalize"] is True

    # check other defaults
    assert vis.config["Arrows"]["opacity"] == 1.0


# def test_config_replace_znsocket(server):
#     room = "test_config_znsocket"
#     vis = ZnDraw(url=server, token=room)

#     vis.config["scene"] = None
#     assert isinstance(vis.config["scene"], znsocket.Dict)


def test_config_modify_scene(server):
    room = "test_config_scene"
    vis = ZnDraw(url=server, token=room)

    vis.config["Camera"]["fps"] = 30
    assert vis.config["Camera"]["fps"] == 30
    vis.config["Camera"]["fps"] = 60
    assert vis.config["Camera"]["fps"] == 60

    # check other defaults
    assert vis.config["Camera"]["camera"] == "PerspectiveCamera"



import pytest
import redis
import znjson
from ase.build import molecule

from zndraw import ZnDraw, ZnDrawLocal
from zndraw.converter import ASEConverter


@pytest.fixture
def full(server):
    return ZnDraw(url=server, token="test_token")


@pytest.fixture
def local(server):
    r = redis.Redis.from_url("redis://localhost:6379/0")

    return ZnDrawLocal(url=server, token="test_token", r=r)


@pytest.mark.parametrize("ref", ["full", "local"])
def test_append_atoms(ref, request):
    """Test the server fixture."""
    vis = request.getfixturevalue(ref)
    water = molecule("H2O")
    vis.append(water)

    assert vis[-1] == water


@pytest.mark.parametrize("ref", ["full", "local"])
def test_append_dump(ref, request):
    """Test the server fixture."""
    vis = request.getfixturevalue(ref)

    water = molecule("H2O")
    vis.append(water)

    assert vis[-1] == water


@pytest.mark.parametrize("ref", ["full", "local"])
def test_append_faulty(ref, request):
    """Test the server fixture."""
    vis = request.getfixturevalue(ref)

    water = molecule("H2O")
    data = ASEConverter().encode(water)
    with pytest.raises(ValueError, match="Unable to parse provided data object"):
        vis.append(data)
    with pytest.raises(ValueError, match="Unable to parse provided data object"):
        vis.extend(3.14)


@pytest.mark.parametrize("ref", ["full", "local"])
def test_setitem_atoms(ref, request, s22):
    """Test the server fixture."""
    vis = request.getfixturevalue(ref)
    vis.extend(s22)
    water = molecule("H2O")
    vis[0] = water
    assert vis[0] == water

    vis[[1, 2]] = [water, water]
    assert vis[[0, 1, 2]] == [water, water, water]


@pytest.mark.parametrize("ref", ["local", "full"])
def test_setitem_dump(ref, request, s22):
    """Test the server fixture."""
    vis = request.getfixturevalue(ref)
    vis.extend(s22)
    water = molecule("H2O")
    vis[0] = water
    assert vis[0] == molecule("H2O")

    vis[[1, 2]] = [water, water]
    assert vis[[0, 1, 2]] == [water, water, water]


@pytest.mark.parametrize("ref", ["full", "local"])
def test_setitem_faulty(ref, request, s22):
    """Test the server fixture."""
    vis = request.getfixturevalue(ref)
    vis.extend(s22)

    water = molecule("H2O")
    data = ASEConverter().encode(water)
    with pytest.raises(ValueError, match="Unable to parse provided data object"):
        vis[0] = data
    with pytest.raises(ValueError, match="Unable to parse provided data object"):
        vis.extend(3.14)


@pytest.mark.parametrize("ref", ["full", "local"])
def test_extend_atoms(ref, request, s22):
    """Test the server fixture."""
    vis = request.getfixturevalue(ref)
    vis.extend(s22)
    assert vis[:] == s22


@pytest.mark.parametrize("ref", ["local", "full"])
def test_extend_dump(ref, request, s22):
    """Test the server fixture."""
    vis = request.getfixturevalue(ref)
    vis.extend(s22)
    assert vis[:] == s22


@pytest.mark.parametrize("ref", ["full", "local"])
def test_extend_faulty(ref, request, s22):
    """Test the server fixture."""
    vis = request.getfixturevalue(ref)
    vis.extend(s22)

    with pytest.raises(ValueError, match="Unable to parse provided data object"):
        vis.extend(znjson.dumps(s22, cls=znjson.ZnEncoder.from_converters(ASEConverter)))

    data = [ASEConverter().encode(s) for s in s22]

    with pytest.raises(ValueError, match="Unable to parse provided data object"):
        vis.extend(data)

    with pytest.raises(ValueError, match="Unable to parse provided data object"):
        vis.extend(3.14)



import numpy as np
import pytest
from ase.build import molecule

from zndraw import ZnDraw


def test_vectors(server):
    vis = ZnDraw(url=server, token="test_token")

    water = molecule("H2O")
    water.info["vectors"] = [[[1, 0, 0], [0, 1, 0]]]

    vis.append(water)

    assert vis[0].info["vectors"] == [[[1, 0, 0], [0, 1, 0]]]


def test_vectors_list_numpy(server):
    vis = ZnDraw(url=server, token="test_token")

    water = molecule("H2O")
    water.info["vectors"] = [np.array([[1, 0, 0], [0, 1, 0]])]

    vis.append(water)

    assert vis[0].info["vectors"] == [[[1, 0, 0], [0, 1, 0]]]


def test_vectors_numpy(server):
    vis = ZnDraw(url=server, token="test_token")

    water = molecule("H2O")
    water.info["vectors"] = np.array([[[1, 0, 0], [0, 1, 0]]])

    vis.append(water)

    assert vis[0].info["vectors"] == [[[1, 0, 0], [0, 1, 0]]]


def test_vectors_format(server):
    vis = ZnDraw(url=server, token="test_token")

    water = molecule("H2O")
    water.info["vectors"] = [1, 2, 3]

    with pytest.raises(ValueError):
        vis.append(water)

    water.info["vectors"] = [[1, 2, 3], [4, 5, 6]]

    with pytest.raises(ValueError):
        vis.append(water)



import json

import ase
import numpy.testing as npt
import pytest
import znjson
from ase.calculators.singlepoint import SinglePointCalculator
from ase.constraints import FixAtoms

from zndraw.converter import ASEConverter


def test_ase_converter(s22):
    s22[0].connectivity = [[0, 1, 1], [1, 2, 1], [2, 3, 1]]
    s22[3].calc = SinglePointCalculator(s22[3])
    s22[3].calc.results = {"energy": 0.0, "predicted_energy": 1.0}
    s22[4].info = {"key": "value"}

    structures_json = znjson.dumps(
        s22, cls=znjson.ZnEncoder.from_converters([ASEConverter])
    )

    non_json = json.loads(structures_json)
    assert "numbers" not in non_json[0]["value"]["arrays"]
    assert "positions" not in non_json[0]["value"]["arrays"]
    assert "pbc" not in non_json[0]["value"]["info"]
    assert "cell" not in non_json[0]["value"]["info"]

    structures = znjson.loads(
        structures_json, cls=znjson.ZnDecoder.from_converters([ASEConverter])
    )
    for s1, s2 in zip(s22, structures):
        assert s1 == s2

    npt.assert_array_equal(structures[0].connectivity, [[0, 1, 1], [1, 2, 1], [2, 3, 1]])
    with pytest.raises(AttributeError):
        _ = structures[1].connectivity

    assert structures[3].calc.results == {"energy": 0.0, "predicted_energy": 1.0}

    assert "colors" not in structures[0].arrays
    assert "radii" not in structures[0].arrays

    assert structures[4].info == {"key": "value"}


def test_exotic_atoms():
    atoms = ase.Atoms("X", positions=[[0, 0, 0]])
    atoms.arrays["colors"] = ["#ff0000"]
    atoms.arrays["radii"] = [0.3]

    new_atoms = znjson.loads(
        znjson.dumps(atoms, cls=znjson.ZnEncoder.from_converters([ASEConverter])),
        cls=znjson.ZnDecoder.from_converters([ASEConverter]),
    )
    npt.assert_array_equal(new_atoms.arrays["colors"], ["#ff0000"])
    npt.assert_array_equal(new_atoms.arrays["radii"], [0.3])


def test_modified_atoms():
    atoms = ase.Atoms("H2", positions=[[0, 0, 0], [0, 0, 1]])
    new_atoms = znjson.loads(
        znjson.dumps(atoms, cls=znjson.ZnEncoder.from_converters([ASEConverter])),
        cls=znjson.ZnDecoder.from_converters([ASEConverter]),
    )
    npt.assert_array_equal(new_atoms.get_atomic_numbers(), [1, 1])

    # subtract
    atoms = new_atoms[:1]
    new_atoms = znjson.loads(
        znjson.dumps(atoms, cls=znjson.ZnEncoder.from_converters([ASEConverter])),
        cls=znjson.ZnDecoder.from_converters([ASEConverter]),
    )

    npt.assert_array_equal(new_atoms.get_atomic_numbers(), [1])

    # add
    atoms = new_atoms + ase.Atoms("H", positions=[[0, 0, 1]])

    new_atoms = znjson.loads(
        znjson.dumps(atoms, cls=znjson.ZnEncoder.from_converters([ASEConverter])),
        cls=znjson.ZnDecoder.from_converters([ASEConverter]),
    )

    npt.assert_array_equal(new_atoms.get_atomic_numbers(), [1, 1])


def test_constraints_fixed_atoms():
    atoms = ase.Atoms("H2", positions=[[0, 0, 0], [0, 0, 1]])
    atoms.set_constraint(FixAtoms([0]))
    new_atoms = znjson.loads(
        znjson.dumps(atoms, cls=znjson.ZnEncoder.from_converters([ASEConverter])),
        cls=znjson.ZnDecoder.from_converters([ASEConverter]),
    )
    assert isinstance(new_atoms.constraints[0], FixAtoms)
    assert new_atoms.constraints[0].index == [0]



import pytest

from zndraw import ZnDraw


def test_bookmarks(server, s22):
    """Test the server fixture."""
    vis = ZnDraw(url=server, token="test_token")
    vis.extend(s22)

    assert vis.bookmarks == {}
    vis.bookmarks = {5: "Hey there!"}
    assert vis.bookmarks == {"5": "Hey there!"}

    vis.bookmarks[5] = "Hello!"
    assert vis.bookmarks[5] == "Hello!"

    with pytest.raises(ValueError):
        vis.bookmarks = ["Bookmarks are not a list!"]

    with pytest.raises(ValueError):
        vis.bookmarks = {5: object()}

    with pytest.raises(ValueError):
        vis.bookmarks = {"string": "Hey there!"}



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import zndraw

project = "ZnDraw"
copyright = "2024, Fabian Zills"
author = "Fabian Zills"
release = zndraw.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "nbsphinx",
    "sphinx_copybutton",
]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]

html_theme_options: dict = {
    "light_logo": "zndraw-light.svg",
    "dark_logo": "zndraw-dark.svg",
    "footer_icons": [
        {
            "name": "GitHub",
            "url": "https://github.com/zincware/zndraw",
            "html": "",
            "class": "fa-brands fa-github fa-2x",
        },
    ],
    "source_repository": "https://github.com/zincware/zndraw",
    "source_branch": "main",
    "source_directory": "docs/source/",
    "navigation_with_keys": True,
}

# font-awesome logos
html_css_files = [
    "https://cdnjs.cloudflare.com/ajax/libs/font-awesome/6.0.0/css/brands.min.css",
]



import numpy as np
import numpy.testing as npt

from zndraw import ZnDraw


def test_points_segments(server):
    """Test the server fixture."""
    zndraw = ZnDraw(url=server, token="test_token")

    assert isinstance(zndraw.points, np.ndarray)
    assert isinstance(zndraw.segments, np.ndarray)

    npt.assert_array_equal(zndraw.points, [])
    npt.assert_array_equal(zndraw.segments, [])

    zndraw.points = np.array([[0, 0, 0], [1, 1, 1]])
    assert zndraw.points.shape == (2, 3)
    assert zndraw.segments.shape == (100, 3)

    npt.assert_array_equal(zndraw.points, [[0, 0, 0], [1, 1, 1]])
